import asyncio
from functools import update_wrapper
from time import time
from typing import Callable, List, Optional, Union
from uuid import uuid4

Seconds = Union[float, int]


class GlobalTask:
    """
    GlobalTask class used to instantiate a task using a function decorator
    """

    retry_attempt = 0
    max_retries: int
    retry_intervals: List[int] = [3, 5, 15]
    last_attempt_time = 0
    delay: Seconds = 0
    func: Optional[Callable] = None
    queue: Optional[asyncio.Queue] = None

    def __init__(
        self,
        func: Callable,
        max_retries: int,
        queue: Optional[asyncio.Queue] = None,
        last_attempt_time=0,
        retry_attempt=0,
        retry_intervals: Optional[List[int]] = None,
        delay: Seconds = 0,
    ):

        self.max_retries = max_retries
        self.retry_intervals = retry_intervals if retry_intervals else self.retry_intervals

        if max_retries < 3:
            self.retry_intervals = self.retry_intervals[:max_retries]
        elif max_retries > 3:
            self.retry_intervals = self.retry_intervals + [
                self.retry_intervals[-1] for _ in range(max_retries - len(self.retry_intervals))
            ]

        self.queue = queue
        self.func = func
        self.retry_attempt = retry_attempt
        self.last_attempt_time = last_attempt_time
        self.delay = delay

    async def enqueue(
        self,
        *args,
        queue: Optional[asyncio.Queue] = None,
        delay: Optional[Seconds] = None,
        id: Optional[str] = None,
        **kwargs,
    ) -> str:
        """
        Enqueue a task adds a self-contained `Task` item into the queue with its own context
        """
        q = queue if queue else self.queue

        if not q:
            raise Exception("No queue specified")

        if delay is None:
            delay = self.delay

        task_item = Task(
            self.func,
            queue=q,
            task_args=args,
            task_kwargs=kwargs,
            retry_attempt=self.retry_attempt,
            last_attempt_time=self.last_attempt_time,
            max_retries=self.max_retries,
            retry_intervals=self.retry_intervals,
            delay=delay,
            id=id,
        )
        await q.put(task_item)
        return task_item.id

    async def __call__(self, *args, **kwargs):
        return await self.func(*args, **kwargs)


class Task(GlobalTask):
    """
    Task class holding context and methods for each individual `Task`
    """

    created_at: float

    def __init__(self, *args, delay: Seconds, **kwargs):
        # generate a new task id if one is not already assigned
        task_id = kwargs.pop("id")
        self.id: str = task_id if task_id else str(uuid4())

        # get task specific args and kwargs and don't pass them to the parent as it doesn't know
        # what to do with it
        task_args = kwargs.pop("task_args")
        task_kwargs = kwargs.pop("task_kwargs")

        super().__init__(*args, **kwargs)

        self.args = task_args
        self.kwargs = task_kwargs
        self.created_at = time()
        self.delay = delay

    async def process(self):
        """
        Process this specific task
        """
        await self.func(*self.args, **self.kwargs)

    def is_ready(self):
        """
        Check if task is ready to be processed.
          * If a task was delayed it's not ready until the time passes.
          * After retrying a `Task` there is a cooldown period
            during which task can not be processed.
        """
        if time() < self.created_at + self.delay:
            return False

        if self.last_attempt_time == 0 or self.retry_attempt == 0:
            return True

        backoff_time = self.retry_intervals[self.retry_attempt - 1]
        if (int(time()) - self.last_attempt_time) <= backoff_time:
            return False

        return True

    def done(self):
        """
        Mark this specific task as complete
        """
        self.queue.task_done()

    async def retry(self):
        """
        Retry this task using the same context data
        """
        self.retry_attempt += 1
        self.last_attempt_time = int(time())
        self.done()
        await self.enqueue(*self.args, **self.kwargs, id=self.id)

    def can_retry(self):
        return self.retry_attempt < self.max_retries


def task(
    queue: Optional[asyncio.Queue] = None,
    max_retries: int = 3,
    retry_intervals=None,
    delay: Seconds = 0,
):
    """
    @task decorator wraps a function into a GlobalTask which can create a self-contained task object
    """

    def wrapper(func):
        task_obj = GlobalTask(
            func,
            queue=queue,
            max_retries=max_retries,
            retry_intervals=retry_intervals,
            delay=delay,
        )

        async def callable_task(*args, **kwargs):
            if kwargs.pop("enqueue", False):
                return await task_obj.enqueue(*args, **kwargs)
            else:
                return await task_obj(*args, **kwargs)

        update_wrapper(callable_task, func)
        callable_task.enqueue = task_obj.enqueue
        return callable_task

    return wrapper



import asyncio
import re
from time import time

import pytest

from kolona import Workers, task, workers
from kolona.constants import RUNTIME_ONEOFF
from kolona.exceptions import RetryTask
from kolona.task import Task


@pytest.mark.asyncio
async def test_with_5_tasks_no_exceptions():
    """
    Add 5 tasks to the queue and create 1 workers, all tasks should be processed.
    """
    queue = asyncio.Queue()

    @task(queue=queue)
    async def runner():
        return True

    for _ in range(5):
        await runner.enqueue()

    workers = Workers(queue, "worker", count=1, runtime=RUNTIME_ONEOFF)

    assert queue.qsize() == 5
    await asyncio.gather(*workers.get())
    assert queue.qsize() == 0


@pytest.mark.asyncio
async def test_with_multiple_workers():
    """
    Add 5 tasks to the queue and create 2 workers, all tasks should be processed.
    """
    queue = asyncio.Queue()

    @task(queue=queue)
    async def runner():
        return True

    for _ in range(5):
        await runner.enqueue()

    workers = Workers(queue, "worker", count=2, runtime=RUNTIME_ONEOFF)

    assert workers.count() == 2
    assert queue.qsize() == 5
    await asyncio.gather(*workers.get())
    assert queue.qsize() == 0


@pytest.mark.asyncio
async def test_with_retries_exceeded(mocker):
    """
    Queue 1 tasks with max_retry set to 2. All retries should fail resulting in:
    - 2 calls to log.info
    - 1 call to log.warning
    """
    queue = asyncio.Queue()

    log_spy = mocker.spy(workers, "log")

    @task(queue=queue, max_retries=2, retry_intervals=[1, 1, 1])
    async def runner():
        raise Exception("Whoopsy")

    await runner.enqueue()

    worker = Workers(queue, "worker", count=1, runtime=RUNTIME_ONEOFF)

    assert queue.qsize() == 1
    await asyncio.gather(*worker.get())

    assert log_spy.info.call_count == 2
    for n, l in enumerate(log_spy.info.mock_calls, start=1):
        assert re.match(
            f"Error occured when processing task runner (.*), retrying: {n}/2", l.args[0]
        )

    assert log_spy.warning.call_count == 1
    for n, l in enumerate(log_spy.warning.mock_calls, start=1):
        assert re.match("Error processing task runner (.*) in last 2 attempts: Whoopsy", l.args[0])

    assert queue.qsize() == 0


@pytest.mark.asyncio
async def test_with_more_than_3_retries_exceeded(mocker):
    """
    Queue 1 tasks with max_retry set to 6. All retries should fail resulting in:
    - 6 calls to log.info
    - 1 call to log.warning
    """
    queue = asyncio.Queue()

    log_spy = mocker.spy(workers, "log")

    @task(queue=queue, max_retries=6, retry_intervals=[1, 1, 1])
    async def runner():
        raise Exception("Whoopsy")

    await runner.enqueue()

    worker = Workers(queue, "worker", count=1, runtime=RUNTIME_ONEOFF)

    assert queue.qsize() == 1
    await asyncio.gather(*worker.get())

    assert log_spy.info.call_count == 6
    for n, l in enumerate(log_spy.info.mock_calls, start=1):
        assert re.match(
            f"Error occured when processing task runner (.*), retrying: {n}/6", l.args[0]
        )

    assert log_spy.warning.call_count == 1
    for n, l in enumerate(log_spy.warning.mock_calls, start=1):
        assert re.match("Error processing task runner (.*) in last 6 attempts: Whoopsy", l.args[0])

    assert queue.qsize() == 0


@pytest.mark.asyncio
async def test_with_retries_ok(mocker):
    """
    Make first process call raise an exception which should trigger a task retry. On the 1st retry
    the task should succeed. This should cause:
    - 1 call to log.info
    """
    queue = asyncio.Queue()

    log_spy = mocker.spy(workers, "log")

    @task(queue=queue, retry_intervals=[1, 1, 1])
    async def runner():
        return True

    # make sure the 1st process call raises an exception
    process_spy = mocker.spy(Task, "process")
    process_spy.side_effect = [Exception("derp"), True]

    for _ in range(1):
        await runner.enqueue()

    worker = Workers(queue, "worker", count=1, runtime=RUNTIME_ONEOFF)

    assert queue.qsize() == 1
    await asyncio.gather(*worker.get())

    assert log_spy.info.call_count == 1
    for n, l in enumerate(log_spy.info.mock_calls, start=1):
        assert re.match(
            f"Error occured when processing task runner (.*), retrying: {n}/3", l.args[0]
        )

    assert log_spy.warning.call_count == 0
    assert queue.qsize() == 0


@pytest.mark.asyncio
async def test_with_manually_triggered_retries(mocker):
    queue = asyncio.Queue()

    log_spy = mocker.spy(workers, "log")

    @task(queue=queue, retry_intervals=[0.1, 0.1, 0.1])
    async def runner(item, name=None):
        raise RetryTask

    task_id = await runner.enqueue(1337, name="Rough")

    worker = Workers(queue, "worker", count=1, runtime=RUNTIME_ONEOFF)

    assert queue.qsize() == 1
    await asyncio.gather(*worker.get())

    assert log_spy.info.call_count == 3
    for n, l in enumerate(log_spy.info.mock_calls, start=1):
        assert re.match(f"Retrying task runner (.*): {n}/3", l.args[0])

    assert log_spy.warning.call_count == 1
    for n, l in enumerate(log_spy.warning.mock_calls, start=1):
        # check if extra kwargs are set as expected
        assert l.kwargs["extra"] == {"task_id": task_id, "arg_0": 1337, "name": "Rough"}

    assert queue.qsize() == 0


@pytest.mark.asyncio
async def test_with_task_args(mocker):
    queue = asyncio.Queue()

    log_spy = mocker.spy(workers, "log")

    @task(queue=queue, retry_intervals=[0.1, 0.1, 0.1])
    async def runner(item, name=None):
        assert item
        assert name is not None
        raise RetryTask

    task_id = await runner.enqueue(1337, name="Rough")

    worker = Workers(queue, "worker", count=1, runtime=RUNTIME_ONEOFF)

    assert queue.qsize() == 1
    await asyncio.gather(*worker.get())

    assert log_spy.info.call_count == 3
    for n, l in enumerate(log_spy.info.mock_calls, start=1):
        assert re.match(f"Retrying task runner (.*): {n}/3", l.args[0])

    assert log_spy.warning.call_count == 1
    for n, l in enumerate(log_spy.warning.mock_calls, start=1):
        # check if extra kwargs are set as expected
        assert l.kwargs["extra"] == {"task_id": task_id, "arg_0": 1337, "name": "Rough"}

    assert queue.qsize() == 0


@pytest.mark.asyncio
async def test_decorated_task_retains_original_props():
    """
    Decorated taks retains __name__, __doc__ and other magic attributes.
    """
    queue = asyncio.Queue()

    @task(queue=queue)
    async def runner():
        """docs"""
        return True

    assert runner.__name__ == "runner"
    assert runner.__doc__ == "docs"


@pytest.mark.asyncio
async def test_delay():
    queue = asyncio.Queue()

    @task(queue=queue)
    async def runner():
        pass

    await runner.enqueue(delay=2)

    worker = Workers(queue, "worker", count=1, runtime=RUNTIME_ONEOFF)

    assert queue.qsize() == 1

    start = time()
    await asyncio.gather(*worker.get())
    stop = time()

    assert stop - start > 2


async def test_id_after_retries(mocker):
    """
    Test if task id is being passed on when it's retried
    """
    queue = asyncio.Queue()

    log_spy = mocker.spy(workers, "log")

    @task(queue=queue, retry_intervals=[1, 1, 1])
    async def runner():
        raise Exception("derp")

    task_ids = []
    for _ in range(2):
        task_ids.append(await runner.enqueue())

    worker = Workers(queue, "worker", count=1, runtime=RUNTIME_ONEOFF)

    assert queue.qsize() == 2
    await asyncio.gather(*worker.get())

    assert log_spy.info.call_count == 6
    for err in log_spy.info.mock_calls:
        errors = [
            f"Error occured when processing task runner ({task_ids[0]}), retrying: 1/3",
            f"Error occured when processing task runner ({task_ids[1]}), retrying: 1/3",
            f"Error occured when processing task runner ({task_ids[0]}), retrying: 2/3",
            f"Error occured when processing task runner ({task_ids[1]}), retrying: 2/3",
            f"Error occured when processing task runner ({task_ids[0]}), retrying: 3/3",
            f"Error occured when processing task runner ({task_ids[1]}), retrying: 3/3",
        ]
        assert err.args[0] in errors

    assert log_spy.warning.call_count == 2
    assert queue.qsize() == 0


async def test_max_retry_zero(mocker):
    """
    Test if task is executed only once in case max_retry is set to 0
    """
    queue = asyncio.Queue()

    log_spy = mocker.spy(workers, "log")

    @task(queue=queue, max_retries=0)
    async def runner():
        raise Exception("derp")

    task_id = await runner.enqueue()

    worker = Workers(queue, "worker", count=1, runtime=RUNTIME_ONEOFF)

    assert queue.qsize() == 1
    await asyncio.gather(*worker.get())

    assert log_spy.warning.call_count == 1
    for err in log_spy.warning.mock_calls:
        assert err.args[0] == f"Error processing task runner ({task_id}) in last 0 attempts: derp"

    assert log_spy.warning.call_count == 1
    assert queue.qsize() == 0


async def test_calling_task_without_enqueing():
    @task()
    async def random_task():
        return True

    result = await random_task()
    assert result is True



import asyncio
from typing import List

from kolona.constants import RUNTIME_INFINITY, RUNTIME_ONEOFF
from kolona.exceptions import RetryTask
from kolona.logger import get_log
from kolona.task import Task

log = get_log("workers")

QUEUE: asyncio.Queue = asyncio.Queue()


class Workers:
    loop = None
    _workers: List[asyncio.tasks.Task] = []
    _queue: asyncio.Queue
    _runtime: int

    def __init__(self, queue: asyncio.Queue, name: str, count=1, runtime=RUNTIME_INFINITY):
        self._queue = queue
        self.loop = asyncio.get_event_loop()

        self._runtime = runtime

        workers = []
        for nr in range(1, 1 + count):
            n = f"{name}-{nr}"
            workers.append(asyncio.create_task(self.processor(n), name=n))
        self._workers = workers

    async def processor(self, name):
        log.debug(f"Starting worker: {name}")

        while True:
            if self._runtime == RUNTIME_ONEOFF and self._queue.empty():
                break

            task: Task = await self._queue.get()
            if not task.is_ready():
                # task is not ready to be processed, add it back to the queue
                await self._queue.put(task)
                continue

            try:
                await task.process()
            except RetryTask:
                if task.can_retry():
                    await task.retry()
                    log.info(
                        f"Retrying task {task.func.__name__} ({task.id}): {task.retry_attempt}/{task.max_retries}",
                    )
                    continue
                else:
                    # task has been retried several times without successfully being processed,
                    # don't process this task anymore
                    task.done()

                    extra_info = {"task_id": task.id}
                    for nr, arg in enumerate(task.args):
                        extra_info[f"arg_{nr}"] = arg

                    if task.kwargs:
                        extra_info.update(task.kwargs)

                    log.warning(
                        f"Failed to process task {task.func.__name__} ({task.id}) in last {task.max_retries} attempts",
                        extra=extra_info,
                    )
                    continue
            except asyncio.CancelledError:
                await task.retry()
            except Exception as e:
                log.exception(e)
                if task.can_retry():
                    await task.retry()
                    log.info(
                        f"Error occured when processing task {task.func.__name__} ({task.id}), retrying: {task.retry_attempt}/{task.max_retries}"
                    )
                    continue
                else:
                    # task has been retried several times without successfully being processed,
                    # don't process this task anymore
                    task.done()

                    extra_info = {"task_id": task.id}
                    for nr, arg in enumerate(task.args):
                        extra_info[f"arg_{nr}"] = arg

                    if task.kwargs:
                        extra_info.update(task.kwargs)

                    log.warning(
                        f"Error processing task {task.func.__name__} ({task.id}) in last {task.max_retries} attempts: {e}",
                        extra=extra_info,
                    )
                    continue
            else:
                # mark task as done when successfully processed
                task.done()

    def get(self):
        return self._workers

    def count(self):
        return len(self._workers)
